/* minesweeper.c
 * shoddy minesweeper clone
 *
 * author: skalar
 * creation/fin date: 5/sep/2021
 *
 * dependencies: libmanstr <attainable from xnopyt>
 *
 * compilation instructions:
 * 		clang minesweeper.c -lmstr
 *
 * please ignore segfaults, am not coderbus
 *
 * controls:
 * 		arrow keys | move cursor
 * 		F		   | flag a tile
 * 		enter	   | unhide a tile
 * 		q		   | quit
 */

/* preprocessor defines */

//#define DEBUG

#define PROJ_NAME "minecweeper"
#define nput(f) (printf("[" PROJ_NAME "] " f))
#define nfputf(f, ...) (printf("[" PROJ_NAME "] " f, __VA_ARGS__))

#define SCR_OF 1

#define OFF_X 5
#define OFF_Y 3

#define BOMB_VAL -2
#define UNKNOWN 10

/* system includes */
#define _GNU_SOURCE
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<termios.h>
#include<time.h>

/* library includes */
#include<manstr.h>

/* local includes */

/* code */
	// types
typedef struct vector2 {
	int x, y;
} v2;

typedef struct cell {
	char val;
	char discovered;
	char flagged;
	char exploded, explodedNow;
} cell_t;

typedef struct cursor {
	v2 loc;
} cursor_t;

typedef struct game {
	v2 size;
	cell_t** field;
	cursor_t cursor;
	time_t startTime;
	unsigned int moves, flags, b;
} game_t;

	// globals
static struct termios G_originalTermios;
static char* G_numColors[] = {
	[1] = "\033[34;49m",
	"\033[36;49m",
	"\033[32;49m",
	"\033[33;49m",
	"\033[31;31m",
};
	// functions

void onExitFunction() {
	printf("\033[?25h");
}

void disableRawMode() {
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &G_originalTermios);
}

void enableRawMode() {
	struct termios raw;
	tcgetattr(STDIN_FILENO, &raw);
	G_originalTermios = raw;
	atexit(disableRawMode);
	raw.c_lflag &= ~(ECHO | ICANON);
	// raw.c_lflag &= ~(ISIG);
	raw.c_cc[VMIN] = 0;
	raw.c_cc[VTIME] = 0;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw);
}

void seedRandom() {
	FILE* randFile = fopen("/dev/urandom", "r");
	int q;
	fread(&q, 1, 4, randFile);
	srand(q);
	fclose(randFile);
}

cell_t** initField(v2 size) {
	cell_t** r_field = calloc(size.y, sizeof(cell_t*));
	for(int i = 0; i < size.y; i++)
		r_field[i] = calloc(size.x, sizeof(cell_t));
	return r_field;
}

void generateField(cell_t** p_field, v2 size, unsigned short numOfBombs) {
	/* generate bombs */	
	for(int i = 0; i <= numOfBombs; i++) {
		int x = rand() % size.x;
		int y = rand() % size.y;
		if(p_field[y][x].val == BOMB_VAL) continue;
		p_field[y][x].val = BOMB_VAL;
		for(int y_scan = y - 1; y_scan <= y + 1; y_scan++) {
			for(int x_scan = x - 1; x_scan <= x + 1; x_scan++) {
				if(x_scan >= 0 && x_scan < size.x && y_scan >= 0 && y_scan < size.y)
					if(p_field[y_scan][x_scan].val != BOMB_VAL)
						p_field[y_scan][x_scan].val++;
			}
		}
	}
}

game_t initGame(v2 p_size, int b) {
	game_t r_game = {0};

	r_game.size = p_size;
	r_game.field = initField(p_size);
	r_game.startTime = time(NULL);

	generateField(r_game.field, r_game.size, b);

	r_game.flags = b;
	r_game.b = b;

#ifdef DEBUG
	for(int y = 0; y < p_size.y; y++)
		for(int x = 0; x < p_size.x; x++)
			r_game.field[y][x].discovered = 1;
#endif

	return r_game;
}

void flushBuf() {
	int c;
	read(STDIN_FILENO, &c, 1);
	while(c != '\n' && c != EOF && read(STDIN_FILENO, &c, 1) != 0);
}

void loss(game_t* p_game) {
	mstr_t frame = mstr("");

	for(int i = 0; i < 25; i++) {
		astrf(&frame,"\033[%i;%iH", p_game->cursor.loc.y + SCR_OF + OFF_Y,\
				p_game->cursor.loc.x + SCR_OF + OFF_X);
		astrf(&frame,"\033[38;5;%sm*", (i % 2 == 0) ? "196" : "124" );
		write(0, str(frame), frame.len);
		fflush(stdout);
		clearString(&frame);
		usleep(50000);
	}

	p_game->field[p_game->cursor.loc.y][p_game->cursor.loc.x].exploded = 1;
	for(;;) {
		int done = 0;
		for(int y = 0; y < p_game->size.y; y++) 
			for(int x = 0; x < p_game->size.x; x++) {
				if(p_game->field[y][x].exploded == 0) done++;
				if(p_game->field[y][x].exploded == 1) {
					for(int u = y + 1; u >= y - 1; u--)
						for(int v = x - 1; v <= x + 1; v++)
							if(u >= 0 && u < p_game->size.y && v >= 0 && v < p_game->size.x\
									&& p_game->field[u][v].explodedNow == 0)
								p_game->field[u][v].explodedNow = 1;
				}
			}
		for(int y = 0; y < p_game->size.y; y++)
			for(int x = 0; x < p_game->size.x; x++)
				if(p_game->field[y][x].exploded == 1 \
						|| p_game->field[y][x].explodedNow == 1) {
					astrf(&frame,"\033[%i;%iH\033[38;5;%im*", y + SCR_OF + OFF_Y, \
						x + SCR_OF + OFF_X, ((y * x) % 2 == 1) ? 196 : 124);
					p_game->field[y][x].explodedNow = 0;
					p_game->field[y][x].exploded = 1;
				}
		write(0, str(frame), frame.len);
		clearString(&frame);
		usleep(7500);
		if(done == 0) break;
	}
	
	for(int i = 0; i < 10; i++) {
		mstr_t frame = mstr("");
		for(int y = 0; y < p_game->size.y; y++)
			for(int x = 0; x < p_game->size.x; x++)
				astrf(&frame,"\033[%i;%iH\033[38;5;%im*", y + SCR_OF + OFF_Y, \
						x + SCR_OF + OFF_X, ((y + x + i) % 2) ? 124 : 196);
		write(0, str(frame), frame.len);
		clearString(&frame);
		usleep(100000);
	}

	astrf(&frame,"\033[%i;%iH\033[0Jkablooey, you fuckin' idiot.\n", 3 + p_game->size.y + \
			SCR_OF + OFF_Y, 5);
	astrf(&frame,"\n\033[;49mtotal moves: \033[32;49m%i", p_game->moves);
	int elapsedTime = time(NULL) - p_game->startTime;
	int sec = elapsedTime % 60, min = elapsedTime / 60;
	astrf(&frame,"\n\033[;49mtotal time: \033[32;49m");
	(min < 10) ? astrf(&frame,"0%i:", min) : astrf(&frame,"%i:", min);
	(sec < 10) ? astrf(&frame,"0%i", sec) : astrf(&frame,"%i", sec);
	astrf(&frame,"\n\033[;49m");

	write(0, str(frame), frame.len);
	freeString(&frame);

	exit(1);
}

void marchZeros(game_t* p_game, v2 cell_pos) {
	cell_t* currentCell = &p_game->field[cell_pos.y][cell_pos.x];
	if(currentCell->val != BOMB_VAL && currentCell->discovered == 0) {
		currentCell->discovered = 1;
		for(int y = cell_pos.y - 1; y <= cell_pos.y + 1; y++)
			for(int x = cell_pos.x - 1; x <= cell_pos.x + 1; x++)
				if(y >= 0 && y < p_game->size.y && x >= 0 && x < p_game->size.x\
						&& currentCell->val == 0)
					marchZeros(p_game, (v2){x, y});
	}
}

void behavoir(game_t* p_game) {
	cell_t* currentCell = &p_game->field[p_game->cursor.loc.y][p_game->cursor.loc.x];
	char buf = 0;
	read(STDIN_FILENO, &buf, 1);
	switch(buf) {
		case '\033':
			read(STDIN_FILENO, &buf, 1);
			read(STDIN_FILENO, &buf, 1);
			switch(buf) {
				case 'A':
					p_game->cursor.loc.y--;
					break;
				case 'B':
					p_game->cursor.loc.y++;
					break;
				case 'C':
					p_game->cursor.loc.x++;
					break;
				case 'D':
					p_game->cursor.loc.x--;
					break;
			}
			break;
		case '\n':
			if(currentCell->flagged == 0) {
				if(currentCell->val == 0)
					marchZeros(p_game, p_game->cursor.loc);
				else if (currentCell->val == BOMB_VAL) {
					if(p_game->moves == 0) {
						generateField(p_game->field, p_game->size, p_game->b); 
						p_game->moves = -1;
					} else loss(p_game);
				} else
					currentCell->discovered = 1;
				p_game->moves++;
			}
			break;
		case 'f':
			if(p_game->flags > 0) {
				if((currentCell->flagged ^= 1) == 1)
					p_game->flags--;
				else p_game->flags++;
			} else {
				if(currentCell->flagged == 1) {
					currentCell->flagged = 0;
					p_game->flags++;
				}
			}
			p_game->moves++;
			break;
		case 'q':
			exit(1);
	}

	// lock cursor to margins
	if(p_game->cursor.loc.x < 0) p_game->cursor.loc.x = p_game->size.x - 1;
	if(p_game->cursor.loc.x >= p_game->size.x) p_game->cursor.loc.x = 0;
	if(p_game->cursor.loc.y < 0) p_game->cursor.loc.y = p_game->size.y - 1;
	if(p_game->cursor.loc.y >= p_game->size.y) p_game->cursor.loc.y = 0;
	flushBuf();
}

void draw(game_t* p_game) {
	mstr_t frame = mstr("");
	astrf(&frame, "\033[2J");

	for(int x = -1; x <= p_game->size.x; x++) {
		astrf(&frame, "\033[33;49m");
		astrf(&frame, "\033[%i;%iH@", SCR_OF + OFF_Y - 1, x + SCR_OF + OFF_X);
		astrf(&frame, "\033[%i;%iH@", SCR_OF + OFF_Y + p_game->size.y, x + SCR_OF + OFF_X);
	}
	

	for(int y = -1; y <= p_game->size.y; y++) {
		astrf(&frame, "\033[%i;%iH@", SCR_OF + OFF_Y + y, SCR_OF + OFF_X - 1);
		astrf(&frame,"\033[%i;%iH@", SCR_OF + OFF_Y + y, SCR_OF + OFF_X + p_game->size.x);
	}
	astrf(&frame,"\033[;49m");

	for(int y = 0; y < p_game->size.y; y++) {
		for(int x = 0; x < p_game->size.x; x++) {
			cell_t* currentCell = &p_game->field[y][x];
			astrf(&frame,"\033[%i;%iH", y + SCR_OF + OFF_Y, x + SCR_OF + OFF_X);

			if(y == p_game->cursor.loc.y && x == p_game->cursor.loc.x)
				astrf(&frame,"\033[7m");

			if(p_game->field[y][x].flagged == 1)
				astrf(&frame,"\033[38;5;242mX");
			else if(p_game->field[y][x].discovered == 0)
				astrf(&frame,".");
			else if (p_game->field[y][x].val == 0)
				astrf(&frame," ");
			else
				astrf(&frame,"%s%i", (currentCell->val < 4) ? \
						G_numColors[(int)currentCell->val] : "", currentCell->val );

			astrf(&frame,"\033[;49m");
		}
	}
	astrf(&frame,"\033[%i;%iHX", p_game->cursor.loc.y + SCR_OF + OFF_Y,\
			p_game->size.x + 2 + SCR_OF + OFF_X);
	astrf(&frame,"\033[%i;%iHX", p_game->size.y + 1 + SCR_OF + OFF_Y,\
			p_game->cursor.loc.x + SCR_OF + OFF_X);
	astrf(&frame,"\033[%i;%iHFLAGS: %i", p_game->size.y + 3 + SCR_OF + OFF_Y,\
			SCR_OF + OFF_X, p_game->flags);	
	astrf(&frame,"\033[%i;%iHMOVES: %i", p_game->size.y + 4 + SCR_OF + OFF_Y,\
			 SCR_OF + OFF_X, p_game->moves);

	int timeElapsed = time(NULL) - p_game->startTime;
	int sec = timeElapsed % 60, min = timeElapsed / 60;

	astrf(&frame,"\033[%i;%iHTIME: ", p_game->size.y + 5 + SCR_OF + OFF_Y,\
			SCR_OF + OFF_X);
	(min < 10) ? astrf(&frame,"0%i:", min) : astrf(&frame,"%i:", min);
	(sec < 10) ? astrf(&frame,"0%i", sec) : astrf(&frame,"%i", sec);

	write(0, str(frame), frame.len);
	freeString(&frame);
}

void won(game_t* p_game) {
	for(int y = 0; y < p_game->size.y; y++)
		for(int x = 0; x < p_game->size.x; x++)
			if(p_game->field[y][x].flagged) {
				printf("\033[%i;%iH\033[32;49mX", y + SCR_OF + OFF_Y,\
						x + SCR_OF + OFF_X);
				fflush(stdout);
				usleep(250000);
			}
	printf("\033[%i;%iH\033[0J u won !", SCR_OF + OFF_Y + p_game->size.y + 2, 0);
	printf("\n\033[;49mtotal moves: \033[32;49m%i", p_game->moves);
	int elapsedTime = time(NULL) - p_game->startTime;
	int sec = elapsedTime % 60, min = elapsedTime / 60;
	printf("\n\033[;49mtotal time: \033[32;49m");
	(min < 10) ? printf("0%i:", min) : printf("%i:", min);
	(sec < 10) ? printf("0%i", sec) : printf("%i", sec);
	printf("\n\033[;49m");
	exit(1);
}

void checkWin(game_t* p_game) {
	int left = 0;
	for(int y = 0; y < p_game->size.y; y++)
		for(int x = 0; x < p_game->size.x; x++)
			if(p_game->field[y][x].val == BOMB_VAL && p_game->field[y][x].flagged == 0)
				left++;
	if(left == 0)
		won(p_game);
}

  
char* inputStr() {
      char* str = calloc(1, 1);
      int len = 0;    
      char buf = '\0';
      for(;(buf = fgetc(stdin)) != 0xa; len++)
      {
          str = realloc(str, len + 2);
          str[len] = buf;
      }
      str[len + 1] = 0;
      return str;
}

int main() { 
	enableRawMode();
	seedRandom();

	int x = 15, y = 15, b = 25;

	for(;;) {
		printf("\033[2J\033[;H");
		char buf = 0;
		read(STDIN_FILENO, &buf, 1);
		printf("[x] %i\n[y] %i\n[b] %i\n[enter] to start\n \
				\033[G[q] to \033[31;49mdie\033[;49m", x, y, b);
		switch(buf) {
			case 'x':
				printf("\033[7;Hx-size: ");
				flushBuf();
				disableRawMode();
				char* q = inputStr();
				int c = atoi(q);
				if(c > 0) x = c;
				enableRawMode();
				break;
			case 'y':
				printf("\033[7;Hy-size: ");
				flushBuf();
				disableRawMode();
				q = inputStr();
				c = atoi(q);
				if(c > 0) y = c;
				if(b > y * x) b = y * x;
				enableRawMode();
				break;
			case 'b':
				printf("\033[7;Hbombs: ");
				flushBuf();
				disableRawMode();
				q = inputStr();
				c = atoi(q);
				if(c > 0 && c < y * x) b = c;
				if(b > y * x) b = y * x;
				enableRawMode();
				break;
			case 'q':
				exit(1);
		}
		if(buf == '\n') break;
		//if(x > 40) x = 40; #fucktherules #yoloswag
		//if(y > 40) y = 40;
		fflush(stdout);
		usleep(25000);
	}

	game_t game = initGame((v2){x, y}, b);

	for(;;) {
		behavoir(&game);
		draw(&game);
		checkWin(&game);
		fflush(stdout);
		usleep(25000);
	}

	disableRawMode();
	return 0;
}
